#!/usr/bin/env python
# coding=utf8

from triflow.core import schemes  # noqa
from triflow.core.model import Model  # noqa
from triflow.core.simulation import Simulation  # noqa

from triflow.plugins.container import TriflowContainer as Container  # noqa
from triflow.plugins.displays import TriflowDisplay as Display  # noqa

import logging
from logging import NullHandler

logging.getLogger(__name__).addHandler(NullHandler())

retrieve_container = Container.retrieve
display_fields = Display.display_fields
display_probe = Display.display_probe



#!/usr/bin/env python
# coding=utf8

import inspect
import logging
import pprint
import time
import warnings
from collections import namedtuple
from uuid import uuid1

import pendulum
import streamz
import tqdm
from numpy import isclose

from . import schemes
from ..plugins.container import TriflowContainer

logging.getLogger(__name__).addHandler(logging.NullHandler())
logging = logging.getLogger(__name__)


def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')


tqdm = tqdm.tqdm_notebook if is_interactive() else tqdm.tqdm


class Timer:
    def __init__(self, last, total):
        self.last = last
        self.total = total

    def __repr__(self):
        repr = """last:   {last}
total:  {total}"""
        return repr.format(last=(pendulum.now()
                                 .subtract(
            seconds=self.last)
            .diff()),
            total=(pendulum.now()
                   .subtract(
                seconds=self.total)
            .diff()))


def null_hook(t, fields, pars):
    return fields, pars


PostProcess = namedtuple(
    "PostProcess", ["name", "function", "description"])


class Simulation(object):
    """High level container used to run simulation build on triflow Model.
      This object is an iterable which will yield every time step until the
      parameters 'tmax' is reached if provided.
      By default, the solver use a 6th order ROW solver, an implicit method
      with integrated time-stepping.

      Parameters
      ----------
      model : triflow.Model
          Contain finite difference approximation and routine of the dynamical
          system
      fields : triflow.BaseFields or dict (any mappable)
          triflow container or mappable filled with initial conditions
      parameters : dict
          physical parameters of the simulation
      dt : float
          time stepping for output. if time_stepping is False, the internal
          time stepping will be the same.
      t : float, optional, default 0.
          initial time
      tmax : float, optional, default None
          Control the end of the simulation. If None (the default), the com-
          putation will continue until interrupted by the user (using Ctrl-C
          or a SIGTERM signal).
      id : None, optional
          Name of the simulation. A 2 word slug will be generated if not
          provided.
      hook : callable, optional, default null_hook.
          Any callable taking the actual time, fields and parameters and
          return modified fields and parameters.
          Will be called every internal time step and can be used to include
          time dependent or conditionnal parameters, boundary conditions...
          The default null_hook has no impact on the computation.
      scheme : callable, optional, default triflow.schemes.RODASPR
          An callable object which take the simulation state and return
          the next step.
          Its signature is scheme.__call__(fields, t, dt, pars, hook)
          and it should return the next time and the updated fields.
          It take the model and extra positional and named arguments.
      time_stepping : boolean, default True
          Indicate if the time step is controlled by an algorithm dependant of
          the temporal scheme (see the doc on time stepping for extra info).
      **kwargs
          extra arguments passed to the scheme.

      Attributes
      ----------
      dt : float
        output time step
      fields : triflow.Fields
        triflow container filled with actual data
      i : int
        actual iteration
      id : str
        name of the simulation
      model : triflow.Model
        triflow Model used in the simulation
      parameters : dict
        physical parameters of the simulation
      status : str
        status of the simulation, one of the following one:
        ('created', 'running', 'finished', 'failed')
      t : float
        actual time
      tmax : float or None, default None
        stopping time of the simulation. Not stopping if set to None.

      Properties
      ----------
      post_processes: list of triflow.core.simulation.PostProcess
        contain all the post processing function attached to the simulation.
      container: triflow.TriflowContainer
        give access to the attached container, if any.
      timer: triflow.core.simulation.Timer
        return the cpu time of the previous step and the total running time of
        the simulation.
      stream: streamz.Stream
        Streamz starting point, fed by the simulation state after each
        time_step. This interface is used for post-processing, saving the data
        on disk by the TriflowContainer and display the fields in real-time.

      Examples
      --------
      >>> import numpy as np
      >>> import triflow
      >>> model = triflow.Model(["k1 * dxxU",
      ...                        "k2 * dxxV"],
      ...                       ["U", "V"],
      ...                       ["k1", "k2"])
      >>> x = np.linspace(0, 100, 1000, endpoint=False)
      >>> U = np.cos(x * 2 * np.pi / 100)
      >>> V = np.sin(x * 2 * np.pi / 100)
      >>> fields = model.fields_template(x=x, U=U, V=V)
      >>> pars = {'k1': 1, 'k2': 1, 'periodic': True}
      >>> simulation = triflow.Simulation(model, fields, pars, dt=5., tmax=50.)
      >>> for t, fields in simulation:
      ...    pass
      >>> print(t)
      50.0
      """  # noqa

    def __init__(self, model, fields, parameters, dt, t=0, tmax=None,
                 id=None, hook=null_hook,
                 scheme=schemes.RODASPR,
                 time_stepping=True, **kwargs):

        def intersection_kwargs(kwargs, function):
            """Inspect the function signature to identify the relevant keys
            in a dictionary of named parameters.
            """
            func_signature = inspect.signature(function)
            func_parameters = func_signature.parameters
            kwargs = {key: value
                      for key, value
                      in kwargs.items() if key in func_parameters}
            return kwargs
        kwargs["time_stepping"] = time_stepping
        self.id = str(uuid1())[:6] if not id else id
        self.model = model
        self.parameters = parameters
        self.fields = model.fields_template(**fields)
        self.t = t
        self.user_dt = self.dt = dt
        self.tmax = tmax
        self.i = 0
        self._stream = streamz.Stream()
        self._pprocesses = []

        self._scheme = scheme(model,
                              **intersection_kwargs(kwargs,
                                                    scheme.__init__))
        if (time_stepping and
            self._scheme not in [schemes.RODASPR,
                                 schemes.ROS3PRL,
                                 schemes.ROS3PRw]):
            self._scheme = schemes.time_stepping(
                self._scheme,
                **intersection_kwargs(kwargs,
                                      schemes.time_stepping))
        self.status = 'created'

        self._total_running = 0
        self._last_running = 0
        self._created_timestamp = pendulum.now()
        self._started_timestamp = None
        self._last_timestamp = None
        self._actual_timestamp = pendulum.now()
        self._hook = hook
        self._container = None
        self._iterator = self.compute()

    def _compute_one_step(self, t, fields, pars):
        """
        Compute one step of the simulation, then update the timers.
        """
        fields, pars = self._hook(t, fields, pars)
        self.dt = (self.tmax - t
                   if self.tmax and (t + self.dt >= self.tmax)
                   else self.dt)
        before_compute = time.clock()
        t, fields = self._scheme(t, fields, self.dt,
                                 pars, hook=self._hook)
        after_compute = time.clock()
        self._last_running = after_compute - before_compute
        self._total_running += self._last_running
        self._last_timestamp = self._actual_timestamp
        self._actual_timestamp = pendulum.now()
        return t, fields, pars

    def compute(self):
        """Generator which yield the actual state of the system every dt.

        Yields
        ------
        tuple : t, fields
            Actual time and updated fields container.
        """
        fields = self.fields
        t = self.t
        pars = self.parameters
        self._started_timestamp = pendulum.now()
        self.stream.emit(self)

        try:
            while True:
                t, fields, pars = self._compute_one_step(t, fields, pars)

                self.i += 1
                self.t = t
                self.fields = fields
                self.parameters = pars
                for pprocess in self.post_processes:
                    pprocess.function(self)
                self.stream.emit(self)
                yield self.t, self.fields

                if self.tmax and (isclose(self.t, self.tmax)):
                    self._end_simulation()
                    return

        except RuntimeError:
            self.status = 'failed'
            raise

    def _end_simulation(self):
        if self.container:
            self.container.flush()
            self.container.merge()

    def run(self, progress=True, verbose=False):
        """Compute all steps of the simulation. Be careful: if tmax is not set,
        this function will result in an infinit loop.

        Returns
        -------

        (t, fields):
            last time and result fields.
        """
        total_iter = int((self.tmax // self.user_dt) if self.tmax else None)
        log = logging.info if verbose else logging.debug
        if progress:
            with tqdm(initial=(self.i if self.i < total_iter else total_iter),
                      total=total_iter) as pbar:
                for t, fields in self:
                    pbar.update(1)
                    log("%s running: t: %g" % (self.id, t))
                try:
                    return t, fields
                except UnboundLocalError:
                    warnings.warn("Simulation already ended")
        for t, fields in self:
            log("%s running: t: %g" % (self.id, t))
        try:
            return t, fields
        except UnboundLocalError:
            warnings.warn("Simulation already ended")

    def __repr__(self):
        repr = """{simulation_name:=^30}

created:      {created_date}
started:      {started_date}
last:         {last_date}

time:         {t:g}
iteration:    {iter:g}

last step:    {step_time}
total time:   {running_time}


Physical parameters
-------------------
{parameters}

Hook function
-------------
{hook_source}

=========== Model ===========
{model_repr}"""
        repr = repr.format(simulation_name=" %s " % self.id,
                           parameters="\n\t".join(
                               [("%s:" % key).ljust(12) +
                                pprint.pformat(value)
                                for key, value
                                in self.parameters.items()]),
                           t=self.t,
                           iter=self.i,
                           model_repr=self.model,
                           hook_source=inspect.getsource(self._hook),
                           step_time=(None if not self._last_running else
                                      pendulum.now()
                                      .subtract(
                                          seconds=self._last_running)
                                      .diff()),
                           running_time=(pendulum.now()
                                         .subtract(
                               seconds=self._total_running)
                               .diff()),
                           created_date=(self._created_timestamp
                                         .to_cookie_string()),
                           started_date=(self._started_timestamp
                                         .to_cookie_string()
                                         if self._started_timestamp
                                         else "None"),
                           last_date=(self._last_timestamp
                                      .to_cookie_string()
                                      if self._last_timestamp
                                      else "None"))
        return repr

    def attach_container(self, path=None, save="all",
                         mode="w", nbuffer=50, force=False):
        """add a Container to the simulation which allows some
        persistance to the simulation.

        Parameters
        ----------
        path : str or None (default: None)
            path for the container. If None (the default), the data lives only
            in memory (and are available with `simulation.container`)
        mode : str, optional
            "a" or "w" (default "w")
        save : str, optional
            "all" will save every time-step,
            "last" will only get the last time step
        nbuffer : int, optional
            wait until nbuffer data in the Queue before save on disk.
        timeout : int, optional
            wait until timeout since last flush before save on disk.
        force : bool, optional (default False)
            if True, remove the target folder if not empty. if False, raise an
            error.
        """
        self._container = TriflowContainer("%s/%s" % (path, self.id)
                                           if path else None,
                                           save=save,
                                           mode=mode, metadata=self.parameters,
                                           force=force, nbuffer=nbuffer)
        self._container.connect(self.stream)
        return self._container

    @property
    def post_processes(self):
        return self._pprocesses

    @property
    def stream(self):
        return self._stream

    @property
    def container(self):
        return self._container

    @property
    def timer(self):
        return Timer(self._last_running, self._total_running)

    def add_post_process(self, name, post_process, description=""):
        """add a post-process

        Parameters
        ----------
        name : str
            name of the post-traitment
        post_process : callback (function of a class with a __call__ method
                                 or a streamz.Stream).
            this callback have to accept the simulation state as parameter
            and return the modifield simulation state.
            if a streamz.Stream is provided, it will me plugged_in with the
            previous streamz (and ultimately to the initial_stream). All these
            stream accept and return the simulation state.
        description : str, optional, Default is "".
            give extra information about the post-processing
        """

        self._pprocesses.append(PostProcess(name=name,
                                            function=post_process,
                                            description=description))
        self._pprocesses[-1].function(self)

    def remove_post_process(self, name):
        """remove a post-process

        Parameters
        ----------
        name : str
            name of the post-process to remove.
        """
        self._pprocesses = [post_process
                            for post_process in self._pprocesses
                            if post_process.name != name]

    def __iter__(self):
        return self.compute()

    def __next__(self):
        return next(self._iterator)



#!/usr/bin/env python
# coding=utf8

import logging
import multiprocessing as mp
import os
import warnings
from collections import deque
from uuid import uuid4

from holoviews import Curve, DynamicMap, Layout, streams
from path import Path  # noqa

log = logging.getLogger(__name__)
log.handlers = []
log.addHandler(logging.NullHandler())


def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')


def manage_display_import():
    global MPLRenderer
    import matplotlib as mpl
    if os.environ.get('DISPLAY', '') == '':
        log.info('no display found. Using non-interactive Agg backend')
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            mpl.use('Agg')

    from holoviews.plotting.mpl import MPLRenderer  # noqa

    if is_interactive():
        from holoviews import notebook_extension
        notebook_extension("bokeh")


manage_display_import()


class TriflowDisplay:
    def __init__(self, skel_data, plot_function,
                 on_disk=None, on_disk_name="triflow_plot",
                 **renderer_args):

        self.on_disk = None
        self._plot_pipe = streams.Pipe(data=skel_data)
        self._dynmap = (DynamicMap(plot_function,
                                   streams=[self._plot_pipe])
                        .opts("Curve [width=600] {+framewise}"))
        self._writers = []
        if on_disk:
            self._renderer = MPLRenderer.instance()
            target_dir = Path(on_disk)
            target_dir.makedirs_p()

            def save_curves(data):
                target_file = target_dir / "%s_%i" % (on_disk_name, data.i)
                process = mp.Process(target=self._renderer.save,
                                     args=(self.hv_curve, target_file),
                                     kwargs=renderer_args)
                self._writers.append(process)
                process.start()

            self._plot_pipe.add_subscriber(save_curves)

    def _repr_mimebundle_(self, *args, **kwargs):
        return self.hv_curve._repr_mimebundle_(*args, **kwargs)

    def connect(self, stream):
        stream.sink(self._plot_pipe.send)

    @property
    def hv_curve(self):
        return self._dynmap.collate()

    def __add__(self, other):
        if isinstance(other, TriflowDisplay):
            return self._dynmap + other._dynmap
        return self._dynmap + other

    def __mul__(self, other):
        if isinstance(other, TriflowDisplay):
            return self._dynmap * other._dynmap
        self._dynmap * other

    @staticmethod
    def display_fields(simul, keys="all",
                       on_disk=None, on_disk_name=None,
                       **renderer_args):
        def plot_function(data):
            nonlocal keys
            curves = []
            keys = data.fields.data_vars if keys == "all" else keys

            for var in keys if not isinstance(keys, str) else [keys]:
                displayed_field = data.fields[var]
                if displayed_field.dims != ('x', ):
                    continue
                curves.append(Curve((displayed_field.squeeze())))
            return Layout(curves).cols(1)

        if on_disk and not on_disk_name:
            on_disk_name = "%s_%s" % (simul.id, "-".join(keys))

        display = TriflowDisplay(simul, plot_function,
                                 on_disk=on_disk,
                                 on_disk_name=on_disk_name,
                                 **renderer_args)
        display.connect(simul.stream)
        return display

    @staticmethod
    def display_probe(simul, function,
                      xlabel=None, ylabel=None, buffer=None,
                      on_disk=None, on_disk_name=None,
                      **renderer_args):
        history = deque([], buffer)
        if not xlabel:
            xlabel = str(uuid4())[:6]
        if not ylabel:
            ylabel = function.__name__
        if ylabel == '<lambda>':
            warnings.warn("Anonymous function used, appending random prefix "
                          "to avoid label confusion")
            ylabel += str(uuid4())[:8]

        def plot_function(data):
            history.append(function(simul))
            return Curve(history, kdims=xlabel, vdims=ylabel)

        if on_disk and not on_disk_name:
            on_disk_name = "simul.id_%s" % ylabel

        display = TriflowDisplay(simul, plot_function,
                                 on_disk=on_disk,
                                 on_disk_name=on_disk_name,
                                 **renderer_args)
        display.connect(simul.stream)
        return display



#!/usr/bin/env python
# coding=utf8

from functools import partial

import numpy as np
from scipy.sparse import csc_matrix
from sympy import lambdify


def theano_compiler(model):
    """Take a triflow model and return optimized theano routines.

    Parameters
    ----------
    model: triflow.Model:
        Model to compile

    Returns
    -------
    (theano function, theano_function):
        Optimized routine that compute the evolution equations and their
        jacobian matrix.
    """
    from theano import tensor as T
    from theano.ifelse import ifelse
    import theano.sparse as ths
    from theano import function

    def th_Min(a, b):
        if isinstance(a, T.TensorVariable) or isinstance(b, T.TensorVariable):
            return T.where(a < b, a, b)
        return min(a, b)

    def th_Max(a, b):
        if isinstance(a, T.TensorVariable) or isinstance(b, T.TensorVariable):
            return T.where(a < b, b, a)
        return max(a, b)

    def th_Heaviside(a):
        if isinstance(a, T.TensorVariable):
            return T.where(a < 0, 1, 1)
        return 0 if a < 0 else 1

    mapargs = {arg: T.vector(arg)
               for arg, sarg
               in zip(model._args, model._symbolic_args)}

    to_feed = mapargs.copy()

    x_th = mapargs['x']
    N = x_th.size
    L = x_th[-1] - x_th[0]
    dx = L / (N - 1)
    to_feed['dx'] = dx

    periodic = T.scalar("periodic", dtype="int32")

    middle_point = int((model._window_range - 1) / 2)

    th_args = [mapargs[key]
               for key
               in [*model._indep_vars,
                   *model._dep_vars,
                   *model._help_funcs,
                   *model._pars]] + [periodic]

    map_extended = {}

    for (varname, discretisation_tree) in \
            model._symb_vars_with_spatial_diff_order.items():
        pad_left, pad_right = model._bounds

        th_arg = mapargs[varname]

        per_extended_var = T.concatenate([th_arg[pad_left:],
                                          th_arg,
                                          th_arg[:pad_right]])
        edge_extended_var = T.concatenate([[th_arg[0]] * middle_point,
                                           th_arg,
                                           [th_arg[-1]] * middle_point])

        extended_var = ifelse(periodic,
                              per_extended_var,
                              edge_extended_var)

        map_extended[varname] = extended_var
        for order in range(pad_left, pad_right + 1):
            if order != 0:
                var = ("{}_{}{}").format(varname,
                                         'm' if order < 0 else 'p',
                                         np.abs(order))
            else:
                var = varname
            new_var = extended_var[order - pad_left:
                                   extended_var.size +
                                   order - pad_right]
            to_feed[var] = new_var

    F = lambdify((model._symbolic_args),
                 expr=model.F_array.tolist(),
                 modules=[T, {"Max": th_Max,
                              "Min": th_Min,
                              "Heaviside": th_Heaviside}])(
        *[to_feed[key]
          for key
          in model._args]
    )

    F = T.concatenate(F, axis=0).reshape((model._nvar, N)).T
    F = T.stack(F).flatten()

    J = lambdify((model._symbolic_args),
                 expr=model.J_array.tolist(),
                 modules=[T, {"Max": th_Max,
                              "Min": th_Min,
                              "Heaviside": th_Heaviside}])(
        *[to_feed[key]
          for key
          in model._args]
    )

    J = [j if j != 0 else T.constant(0.)
         for j in J]
    J = [j if not isinstance(j, (int, float)) else T.constant(j)
         for j in J]
    J = T.stack([T.repeat(j, N) if j.ndim == 0 else j
                 for j in J])
    J = J[model._sparse_indices[0]].T.squeeze()

    i = T.arange(N).dimshuffle([0, 'x'])
    idx = T.arange(N * model._nvar).reshape((N, model._nvar)).T
    edge_extended_idx = T.concatenate([T.repeat(idx[:, :1],
                                                middle_point,
                                                axis=1),
                                       idx,
                                       T.repeat(idx[:, -1:],
                                                middle_point,
                                                axis=1)],
                                      axis=1).T.flatten()
    per_extended_idx = T.concatenate([idx[:, -middle_point:],
                                      idx,
                                      idx[:, :middle_point]],
                                     axis=1).T.flatten()
    extended_idx = ifelse(periodic,
                          per_extended_idx,
                          edge_extended_idx)

    rows = T.tile(T.arange(model._nvar),
                  model._window_range * model._nvar) + i * model._nvar
    cols = T.repeat(T.arange(model._window_range * model._nvar),
                    model._nvar) + i * model._nvar
    rows = rows[:, model._sparse_indices].reshape(J.shape).flatten()
    cols = extended_idx[cols][:, model._sparse_indices] \
        .reshape(J.shape).flatten()

    permutation = T.argsort(cols)

    J = J.flatten()[permutation]
    rows = rows[permutation]
    cols = cols[permutation]
    count = T.zeros((N * model._nvar + 1,), dtype=int)
    uq, cnt = T.extra_ops.Unique(False, False, True)(cols)
    count = T.set_subtensor(count[uq + 1], cnt)

    indptr = T.cumsum(count)
    shape = T.stack([N * model._nvar, N * model._nvar])
    sparse_J = ths.CSC(J, rows, indptr, shape)
    F_theano_function = function(inputs=th_args,
                                 outputs=F,
                                 on_unused_input='ignore',
                                 allow_input_downcast=True)
    J_theano_function = function(inputs=th_args,
                                 outputs=sparse_J,
                                 on_unused_input='ignore',
                                 allow_input_downcast=True)

    return F_theano_function, J_theano_function


def numpy_compiler(model):
    """Take a triflow model and return optimized numpy routines.

    Parameters
    ----------
    model: triflow.Model:
        Model to compile

    Returns
    -------
    (numpy function, numpy function):
        Optimized routine that compute the evolution equations and their
        jacobian matrix.
    """

    def np_Min(args):
        a, b = args
        return np.where(a < b, a, b)

    def np_Max(args):
        a, b = args
        return np.where(a < b, b, a)

    def np_Heaviside(a):
        return np.where(a < 0, 1, 1)

    f_func = lambdify((model._symbolic_args),
                      expr=model.F_array.tolist(),
                      modules=[{"amax": np_Max,
                                "amin": np_Min,
                                "Heaviside": np_Heaviside},
                               "numpy"])

    j_func = lambdify((model._symbolic_args),
                      expr=model._J_sparse_array.tolist(),
                      modules=[{"amax": np_Max,
                                "amin": np_Min,
                                "Heaviside": np_Heaviside},
                               "numpy"])

    compute_F = partial(compute_F_numpy, model, f_func)
    compute_J = partial(compute_J_numpy, model, j_func)

    return compute_F, compute_J


def init_computation_numpy(model, *input_args):
    mapargs = {key: input_args[i]
               for i, key
               in enumerate([*model._indep_vars,
                             *model._dep_vars,
                             *model._help_funcs,
                             *[*model._pars, "periodic"]])}
    x = mapargs["x"]
    N = x.size
    L = x[-1] - x[0]
    dx = L / (N - 1)
    periodic = mapargs["periodic"]
    middle_point = int((model._window_range - 1) / 2)

    args = [mapargs[key]
            for key
            in [*model._indep_vars,
                *model._dep_vars,
                *model._help_funcs,
                *model._pars]] + [periodic]

    mapargs['dx'] = dx

    map_extended = mapargs.copy()

    for (varname, discretisation_tree) in \
            model._symb_vars_with_spatial_diff_order.items():
        pad_left, pad_right = model._bounds

        arg = mapargs[varname]
        if periodic:
            extended_var = np.concatenate([arg[pad_left:],
                                           arg,
                                           arg[:pad_right]])
        else:
            extended_var = np.concatenate([[arg[0]] * middle_point,
                                           arg,
                                           [arg[-1]] * middle_point])

        map_extended[varname] = extended_var
        for order in range(pad_left, pad_right + 1):
            if order != 0:
                var = ("{}_{}{}").format(varname,
                                         'm' if order < 0 else 'p',
                                         np.abs(order))
            else:
                var = varname
            new_var = extended_var[order - pad_left:
                                   extended_var.size +
                                   order - pad_right]
            map_extended[var] = new_var
    return args, map_extended, N, middle_point, periodic


def compute_F_numpy(model, f_func, *input_args):
    args, map_extended, N, middle_point, periodic = \
        init_computation_numpy(model, *input_args)
    F = f_func(*[map_extended[key]
                 for key
                 in model._args])
    F = np.concatenate(F, axis=0).reshape((model._nvar, N)).T
    F = np.stack(F).flatten()
    return F


def compute_J_numpy(model, j_func, *input_args):
    args, map_extended, N, middle_point, periodic = \
        init_computation_numpy(model, *input_args)
    J = j_func(*[map_extended[key]
                 for key
                 in model._args])

    J = np.stack([np.repeat(j, N) if len(
        np.array(j).shape) == 0 else j for j in J])
    J = J.T.squeeze()

    i = np.arange(N)[:, None]
    idx = np.arange(N * model._nvar).reshape((N, model._nvar)).T

    if periodic:
        extended_idx = np.concatenate([idx[:, -middle_point:],
                                       idx,
                                       idx[:, :middle_point]],
                                      axis=1).T.flatten()
    else:
        extended_idx = np.concatenate([np.repeat(idx[:, :1],
                                                 middle_point,
                                                 axis=1),
                                       idx,
                                       np.repeat(idx[:, -1:],
                                                 middle_point,
                                                 axis=1)],
                                      axis=1).T.flatten()

    rows = np.tile(np.arange(model._nvar),
                   model._window_range * model._nvar) + i * model._nvar
    cols = np.repeat(np.arange(model._window_range * model._nvar),
                     model._nvar) + i * model._nvar
    rows = rows[:, model._sparse_indices].reshape(J.shape)
    cols = extended_idx[cols][:, model._sparse_indices].reshape(J.shape)
    rows = rows.flatten()
    cols = cols.flatten()

    sparse_J = csc_matrix((J.flatten(), (rows, cols)),
                          shape=(N * model._nvar, N * model._nvar))
    return sparse_J



#!/usr/bin/env python
# coding=utf8

from copy import copy, deepcopy

import numpy as np
import pandas as pd
from xarray import Dataset


def reduce_fields(coords,
                  dependent_variables,
                  helper_functions,
                  data):
    Field = BaseFields.factory(coords,
                               dependent_variables,
                               helper_functions)
    return Field(**data)


class BaseFields(Dataset):
    """Specialized container which expose the data as a structured numpy array,
      give access to the dependants variables and the herlpers function as
      attributes (as a numpy rec array) and is able to give access to a flat
      view of the dependent variables only (which is needed by the ode
      solvers for all the linear algebra manipulation).

      Parameters
      ----------
      **inputs : numpy.array
          named argument providing x, the dependent variables and the helper functions. All of these are mendatory and a KeyError will be raised if a data is missing.

      Attributes
      ----------
      array : numpy.array
        vanilla numpy array containing the data
      size : int
        Number of discretisation nodes
      """  # noqa
    @staticmethod
    def factory(coords,
                dependent_variables,
                helper_functions):
        """Fields factory generating specialized container build around a
          triflow Model and xarray.

          Parameters
          ----------
          coords: iterable of str:
              coordinates name. First coordinate have to be shared with all
              variables
          dependent_variables : iterable tuple (name, coords)
              coordinates and name of the dependent variables
          helper_functions : iterable tuple (name, coords)
              coordinates and name of the helpers functions

          Returns
          -------
          triflow.BaseFields
              Specialized container which expose the data as a structured
              numpy array
          """
        Field = type('Field', BaseFields.__bases__,
                     dict(BaseFields.__dict__))
        Field._coords = coords
        Field.dependent_variables_info = dependent_variables
        Field.helper_functions_info = helper_functions
        Field._var_info = [*list(Field.dependent_variables_info),
                           *list(Field.helper_functions_info)]
        Field.dependent_variables = [dep[0]
                                     for dep
                                     in Field.dependent_variables_info]
        Field.helper_functions = [dep[0]
                                  for dep
                                  in Field.helper_functions_info]
        Field._keys, Field._coords_info = zip(*Field._var_info)
        return Field

    @staticmethod
    def factory1D(dependent_variables,
                  helper_functions):
        """Fields factory generating specialized container build around a
          triflow Model and xarray.
          Wrapper for 1D data.

          Parameters
          ----------
          dependent_variables : iterable for str
              name of the dependent variables
          helper_functions : iterable of str
              name of the helpers functions

          Returns
          -------
          triflow.BaseFields
              Specialized container which expose the data as a structured
              numpy array
          """
        return BaseFields.factory(("x", ),
                                  [(name, ("x", ))
                                   for name
                                   in dependent_variables],
                                  [(name, ("x", ))
                                   for name
                                   in helper_functions],)

    def __init__(self, **inputs):
        Dataset.__init__(self,
                         data_vars={key: (coords, inputs[key])
                                    for key, coords in self._var_info},
                         coords={coord: inputs[coord]
                                 for coord in self._coords})

    def __reduce__(self):
        return (reduce_fields,
                (self._coords,
                 self.dependent_variables_info,
                 self.helper_functions_info,
                 {key: self[key]
                  for key in self.keys()}))

    def copy(self, deep=True):
        new_dataset = Dataset.copy(self, deep)
        new_dataset.__dict__.update({key: (deepcopy(value)
                                           if deep
                                           else copy(value))
                                     for key, value
                                     in self.__dict__.items()})
        return new_dataset

    def __copy__(self, deep=True):
        return self.copy(deep)

    @property
    def size(self):
        """numpy.ndarray.view: view of the dependent variables of the main numpy array
        """  # noqa
        return self["x"].size

    @property
    def uarray(self):
        """numpy.ndarray.view: view of the dependent variables of the main numpy array
        """  # noqa
        return self[self.dependent_variables]

    @property
    def uflat(self):
        """return a flatten **copy** of the main numpy array with only the
        dependant variables.

        Be carefull, modification of these data will not be reflected on
        the main array!
        """  # noqa
        aligned_arrays = [self[key].values[[(slice(None)
                                             if c in coords
                                             else None)
                                            for c in self._coords]].T
                          for key, coords in self.dependent_variables_info]
        return np.vstack(aligned_arrays).flatten("F")

    def keys(self):
        return [*self._coords, *self._keys]

    def to_df(self):
        if len(self.coords) > 1:
            raise ValueError("CSV files only available for 1D arrays")
        data = {key: self[key]
                for key
                in self._keys}
        df = pd.DataFrame(dict(**data), index=self[self._coords[0]])
        return df

    def fill(self, uflat):
        rarray = uflat.reshape((self[self._coords[0]].size,
                                -1))
        ptr = 0
        for var, coords in self.dependent_variables_info:
            coords = list(coords)
            coords.remove(self._coords[0])
            next_ptr = ptr + sum([self.coords[key].size for key in coords])
            next_ptr = ptr + 1 if next_ptr == ptr else next_ptr
            self[var][:] = rarray[:, ptr:next_ptr].squeeze()
            ptr = next_ptr

    def to_csv(self, path):
        self.to_df().to_csv(path)

    def to_clipboard(self):
        self.to_df().to_clipboard()



#!/usr/bin/env python
# coding=utf8

import json
import logging
import warnings
from collections import deque, namedtuple
from uuid import uuid1

import yaml
from path import Path
from streamz import collect
from xarray import concat, open_dataset, open_mfdataset

log = logging.getLogger(__name__)
log.handlers = []
log.addHandler(logging.NullHandler())

FieldsData = namedtuple("FieldsData", ["data", "metadata"])


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def coerce_attr(key, value):
    value_type = type(value)
    if value_type in [int, float, str]:
        return value
    for cast in (int, float, str):
        try:
            value = cast(value)
            log.debug("Illegal netCDF type ({}) of attribute for {}, "
                      "casted to {}".format(value_type, key, cast))
            return value
        except TypeError:
            pass
    raise TypeError("Illegal netCDF type ({}) of attribute for {}, "
                    "auto-casting failed, tried to cast to "
                    "int, float and str")


class TriflowContainer:
    def __init__(self, path=None, mode='a', *, save="all",
                 metadata={}, force=False, nbuffer=50):
        self._nbuffer = nbuffer
        self._mode = mode
        self._metadata = metadata
        self.save = save
        self._cached_data = deque([], self._n_save)
        self._collector = None
        self.path = path = Path(path).abspath() if path else None

        if not path:
            return

        if self._mode == "w" and force:
            path.rmtree_p()

        if self._mode == "w" and not force and path.exists():
            raise FileExistsError(
                "Directory %s exists, set force=True to override it" % path)

        if self._mode == "r" and not path.exists():
            raise FileNotFoundError("Container not found.")
        path.makedirs_p()

        with open(self.path / 'metadata.yml', 'w') as yaml_file:
            yaml.dump(self._metadata,
                      yaml_file, default_flow_style=False)

    @property
    def save(self):
        return "last" if self._n_save else "all"

    @save.setter
    def save(self, value):
        if value == "all":
            self._n_save = None
        elif value == "last" or value == -1:
            self._n_save = 1
        else:
            raise ValueError('save argument accept only "all", "last" or -1 '
                             'as value, not %s' % value)

    def _expand_fields(self, t, fields):
        fields = fields.assign_coords(t=t).expand_dims("t")
        for key, value in self._metadata.items():
            fields.attrs[key] = coerce_attr(key, value)
        self._cached_data.append(fields)
        return fields

    def _concat_fields(self, fields):
        if fields:
            return concat(fields, dim="t")

    def connect(self, stream):
        def get_t_fields(simul):
            return simul.t, simul.fields

        def expand_fields(inps):
            return self._expand_fields(*inps)

        def get_last(list_fields):
            return list_fields[-1]

        accumulation_stream = (stream
                               .map(get_t_fields)
                               .map(expand_fields))

        self._collector = collect(accumulation_stream)
        if self.save == "all":
            self._collector.map(self._concat_fields).sink(self._write)
        else:
            self._collector.map(get_last).sink(self._write)

        (accumulation_stream
         .partition(self._nbuffer)
         .sink(self._collector.flush))

        return self._collector

    def flush(self):
        if self._collector:
            self._collector.flush()

    def _write(self, concatenated_fields):
        if concatenated_fields is not None and self.path:
            target_file = self.path / "data_%i.nc" % uuid1()
            concatenated_fields.to_netcdf(target_file)
            self._cached_data = deque([], self._n_save)
            if self.save == "last":
                [file.remove()
                 for file in self.path.glob("data_*.nc")
                 if file != target_file]

    def __repr__(self):
        repr = """path:   {path}
{data}""".format(path=self.path, data=self.data)
        return repr

    def __del__(self):
        self.flush()

    @property
    def data(self):
        try:
            if self.path:
                return open_mfdataset(self.path / "data*.nc")
            return self._concat_fields(self._cached_data)
        except OSError:
            return

    @property
    def metadata(self):
        try:
            if self.path:
                with open(self.path / 'metadata.yml', 'r') as yaml_file:
                    return yaml.load(yaml_file)
            return self._metadata
        except OSError:
            return

    @metadata.setter
    def metadata(self, parameters):
        if self._mode == "r":
            return
        for key, value in parameters.items():
            self._metadata[key] = value
        if self.path:
            with open(self.path / 'info.yml', 'w') as yaml_file:
                yaml.dump(self._metadata,
                          yaml_file, default_flow_style=False)

    @staticmethod
    def retrieve(path, isel='all', lazy=True):
        path = Path(path)
        try:
            data = open_dataset(path / "data.nc")
            lazy = True
        except FileNotFoundError:
            data = open_mfdataset(path / "data*.nc",
                                  concat_dim="t").sortby("t")
        try:
            with open(Path(path) / 'metadata.yml', 'r') as yaml_file:
                metadata = yaml.load(yaml_file)
        except FileNotFoundError:
            # Ensure retro-compatibility with older version
            with open(path.glob("Treant.*.json")[0]) as f:
                metadata = json.load(f)["categories"]

        if isel == 'last':
            data = data.isel(t=-1)
        elif isel == 'all':
            pass
        elif isinstance(isel, dict):
            data = data.isel(**isel)
        else:
            data = data.isel(t=isel)

        if not lazy:
            return FieldsData(data=data.load(),
                              metadata=AttrDict(**metadata))

        return FieldsData(data=data,
                          metadata=AttrDict(**metadata))

    @staticmethod
    def get_last(path):
        warnings.warn(
            "get_last method is deprecied and will be removed in a "
            "future version. Please use retrieve(path, 'last') instead.",
            DeprecationWarning
        )

        return TriflowContainer.retrieve(path, isel=[-1], lazy=False)

    @staticmethod
    def get_all(path):
        warnings.warn(
            "get_last method is deprecied and will be removed in a "
            "future version. Please use retrieve(path) instead.",
            DeprecationWarning
        )

        return TriflowContainer.retrieve(path, isel="all", lazy=False)

    def merge(self, override=True):
        if self.path:
            return TriflowContainer.merge_datafiles(self.path,
                                                    override=override)

    @staticmethod
    def merge_datafiles(path, override=False):
        path = Path(path)

        if (path / "data.nc").exists() and not override:
            raise FileExistsError(path / "data.nc")
        (path / "data.nc").remove_p()

        split_data = open_mfdataset(path / "data*.nc",
                                    concat_dim="t").sortby("t")
        split_data.to_netcdf(path / "data.nc")
        merged_data = open_dataset(path / "data.nc", chunks=split_data.chunks)

        if not split_data.equals(merged_data):
            (path / "data.nc").remove()
            raise IOError("Unable to merge data ")

        [file.remove() for file in path.files("data_*.nc")]
        return path / "data.nc"



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['triflow', 'triflow.core', 'triflow.plugins']

package_data = \
{'': ['*']}

install_requires = \
['bokeh>=0.12,<0.13',
 'dask>=0.17.2,<0.18.0',
 'holoviews>=1.1,<2.0',
 'numpy>=1,<2',
 'path.py>=11.0,<12.0',
 'pendulum>=1.5,<2.0',
 'scipy>=1,<2',
 'streamz>=0.3,<0.4',
 'sympy>=1,<2',
 'theano>=1,<2',
 'tqdm>=4.23,<5.0',
 'xarray>=0.10,<0.11']

setup_kwargs = {
    'name': 'triflow',
    'version': '0.5.2',
    'description': 'Automatic finite difference discretization for 1D PDE with fast temporal solvers.',
    'long_description': '# Triflow\n\n[![DOI](https://zenodo.org/badge/71994260.svg)](https://zenodo.org/badge/latestdoi/71994260)\n\n|   Master   |  Dev   |\n|:----------:|:------:|\n|[![Build Status](https://travis-ci.org/locie/triflow.svg?branch=master)](https://travis-ci.org/locie/triflow) | [![Build Status](https://travis-ci.org/locie/triflow.svg?branch=dev)](https://travis-ci.org/locie/triflow)\n[![Coverage Status](https://coveralls.io/repos/github/locie/triflow/badge.svg?branch=dev)](https://coveralls.io/github/locie/triflow?branch=master) | [![Coverage Status](https://coveralls.io/repos/github/locie/triflow/badge.svg?branch=dev)](https://coveralls.io/github/locie/triflow?branch=dev)\n\n## Installation\n\n### External requirements\n\nThis library is written for python &gt;= 3.5.\n\nThe library is based on Theano, thus extra dependencies like fortran and\nC compiler are needed, see Theano install page for extra information:\n\n<http://deeplearning.net/software/theano/install.html>\n\nOn v0.5.0, it is possible to choose between theano and numpy (which provide similar features). numpy will be slower but with no compilation time, which is handy for testing and prototyping.\n\n### via PyPI\n\n``` {.sourceCode .bash}\npip install triflow\n```\n\nwill install the package and\n\n``` {.sourceCode .bash}\npip install triflow --upgrade\n```\n\nwill update an old version of the library.\n\nuse sudo if needed, and the user flag if you want to install it without\nthe root privileges:\n\n``` {.sourceCode .bash}\npip install --user triflow\n```\n\n### via github\n\nYou can install the last version of the library using pip and the github\nrepository:\n\n``` {.sourceCode .bash}\npip install git+git://github.com/locie/triflow.git\n```\n\n### via anaconda (strongly advised for windows users)\n\n``` {.sourceCode .bash}\nconda install -c conda-forge -c celliern triflow\n```\n\n## Introduction\n\n### Motivation\n\nThe aim of this library is to have a (relatively) easy way to write\ntransient dynamic systems with 1D finite difference discretization, with\nfast temporal solvers.\n\nThe main two parts of the library are:\n\n- symbolic tools defining the spatial discretization, with boundary\n    taking into account in a separated part\n- a fast temporal solver written in order to use the sparsity of the\n    finite difference method to reduce the memory and CPU usage during\n    the solving\n\nMoreover, extra tools are provided and the library is written in a\nmodular way, allowing an easy extension of these different parts (see\nthe plug-in module of the library.)\n\nThe library fits well with an interactive usage (in a jupyter notebook).\nThe dependency list is actually larger, but on-going work target a\nreduction of the stack complexity.\n\n### Model writing\n\nAll the models are written as function generating the F vector and the\nJacobian matrix of the model defined as `dtU = F(U)`.\n\nThe symbolic model is written as a simple mathematic equation. For\nexample, a diffusion advection model can be written as:\n\n``` {.sourceCode .python}\nfrom triflow import Model\n\nequation_diff = "k * dxxU - c * dxU"\ndependent_var = "U"\nphysical_parameters = ["k", "c"]\n\nmodel = Model(equation_diff, dependent_var, physical_parameters)\n```\n\n### Example\n\n``` {.sourceCode .python}\nimport numpy as np\nimport pylab as pl\nfrom triflow import Model, Simulation\n\nmodel = Model("k * dxxU - c * dxU",\n              "U", ["k", "c"])\n\nx, dx = np.linspace(0, 1, 200, retstep=True)\nU = np.cos(2 * np.pi * x * 5)\nfields = model.fields_template(x=x, U=U)\n\nparameters = dict(c=.03, k=.001, dx=dx, periodic=False)\n\nt = 0\ndt = 5E-1\ntmax = 2.5\n\npl.plot(fields.x, fields.U, label=f\'t: {t:g}\')\n\n\ndef dirichlet_condition(t, fields, pars):\n    fields.U[0] = 1\n    fields.U[-1] = 0\n    return fields, pars\n\n\nsimul = Simulation(model, t, fields, parameters, dt,\n                   hook=dirichlet_condition, tmax=tmax)\n\nfor i, (t, fields) in enumerate(simul):\n    print(f"iteration: {i}\\t",\n          f"t: {t:g}", end=\'\\r\')\n    pl.plot(fields.x, fields.U, label=f\'t: {t:g}\')\n\npl.xlim(0, 1)\nlegend = pl.legend(loc=\'best\')\n\npl.show()\n```\n\n### NEWS\n\nv0.5.0:\n\n- WARNING: some part of the API has changed:\n  - Simulation signature has changed. `t` arg is now optional (with t=0) as default and `physical_parameters` is now `parameters`.\n  - The displays have been completely rewritten, and the previous API is depreciated. Users are encouraged to modify their scripts or to stick to the ^0.4 triflow versions.\n- move schemes from plugins to core\n- compilers: remove tensorflow, add numpy which is way slower but has no compilation overhead.\n- displays and containers are connected to the simulation via `streamz`\n- add post-processing.\n- real-time display is now based on [Holoviews](https://holoviews.org/). Backward compatibility for display is broken and users are encouraged to modify their scripts or to stick to the ^0.4 triflow versions.\n- use poetry to manage dependencies.\n- use `tqdm` to display simulation update.\n\nv0.4.12:\n\n- give user choice of compiler\n  - get out tensorflow compiler (not really efficient, lot of maintenance trouble)\n  - give access to theano and numpy compiler\n- upwind scheme support\n- using xarray as fields backend, allowing easy post process and save\n- update display and containers\n- adding repr string to all major classes\n\nv0.4.7:\n\n- adding tensor flow support with full testing\n- adding post-processing in bokeh fields display\n\n### ROADMAP / TODO LIST\n\nThe following items are linked to a better use of solid external libs:\n\n- change all the display and container workflow:\n  - use streamz to allow pipeline like way to add display / probing / post-process :done:\n  - use holoviews as main way to do real-time plotting :done:\n  - use xarray multi netcdf files to reduce IO lack of performance :done:\n- better use of external solving lib:\n  - merge triflow.plugins.schemes and scipy.integrate.OdeSolver API :todo:\n  - use scipy.integrate.solve_ivp for triflow temporal scheme solving (making it more robust) :todo:\n  - main goal is to have better two-way integration with scipy :todo:\n\nThese are linked to the triflow core\n\n- build a robust boundary condition API :todo:\n- work on dimension extension, allowing 2D resolution and more :todo:\n- allow auxiliary function to make some complex model easier to write :todo:\n- allow a choice on the finite difference scheme, on a global way or term by term :todo:\n- test and propose other compilers (Cython, numba, pythran?) :todo:\n- work on adaptive spatial and temporal mesh :todo:\n\nThese are far away but can be very interesting:\n\n- implement continuation algorithm working with triflow (separate project?)\n- try other kind of discretization scheme (separate project each?)\n  - Finite volume\n  - Finite element?\n\nThe final (and very ambitious goal) is to provide a robust framework allowing\nto link the mathematical language (and a natural way to write the model,\nas natural as possible) with a high performance and robust solving of\nthe numerical system. There is a trade-off between the ease of use and the\nperformance, for this software, all mandatory dependencies have to be\npip-installable (as opposite to dedalus project or fenics project), even if\nsome non-mandatory dependencies can be harder to install.\n\nIf we go that further, it may be interesting to split the project with the\ntriflow language, the different spatial discretization and so on...\n\n### License\n\nThis project is licensed under the term of the [MIT license](LICENSE)\n\n[![image](https://zenodo.org/badge/DOI/10.5281/zenodo.584101.svg)](https://doi.org/10.5281/zenodo.584101)\n',
    'author': 'Nicolas Cellier',
    'author_email': 'contact@nicolas-cellier.net',
    'url': 'http://github.com/locie/triflow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.5',
}


setup(**setup_kwargs)
